'''
Created on 15.02.2018

@author: abaktheer

Microplane damage model 2D - ODFS (Wu [2009])
'''

from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from traits.api import \
    Constant,\
    Property, cached_property
from bmcs_utils.api import Float
import numpy as np
import traits.api as tr


class MATS3DMplDamageODF(MATS3DEval):

    epsilon_0 = Float(59.0e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    epsilon_f = Float(250.0e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    c_T = Float(1.0,
                label="a",
                desc="Lateral pressure coefficient",
                enter_set=True,
                auto_set=False)

    zeta_G = Float(1.0,
                   label="zeta_G",
                   desc="anisotropy parameter",
                   enter_set=True,
                   auto_set=False)

    E = tr.Float(34000.0,
                 label="E",
                 desc="Young's Modulus",
                 auto_set=False,
                 input=True)

    nu = tr.Float(0.2,
                  label='nu',
                  desc="Poison ratio",
                  auto_set=False,
                  input=True)

    state_var_shapes = tr.Property(tr.Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''
    @cached_property
    def _get_state_var_shapes(self):
        return {'kappa': (self.n_mp,),
                'omega': (self.n_mp,)}

    #-------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    #-------------------------------------------------------------------------

    # get the dyadic product of the microplane normals
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        # dyadic product of the microplane normals

        MPNN_nij = np.einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')

    @cached_property
    def _get__MPTT(self):
        # Third order tangential tensor for each microplane
        delta = np.identity(3)
        MPTT_nijr = 0.5 * (np.einsum('ni,jr -> nijr', self._MPN, delta) +
                           np.einsum('nj,ir -> njir', self._MPN, delta) - 2.0 *
                           np.einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN))
        return MPTT_nijr

    def _get_e_Emna(self, eps_Emab):
        # Projection of apparent strain onto the individual microplanes
        e_ni = np.einsum('nb,Emba->Emna', self._MPN, eps_Emab)
        return e_ni

    def _get_e_N_Emn(self, e_Emna):
        # get the normal strain array for each microplane
        e_N_Emn = np.einsum('Emna, na->Emn', e_Emna, self._MPN)
        return e_N_Emn

    def _get_e_N_arr_2(self, eps_Emab):

        #eps_mtx = self.map_eps_eng_to_mtx(eps_eng)
        return np.einsum('nij,Emij->Emn', self._MPNN, eps_Emab)

    def _get_e_t_vct_arr_2(self, eps_Emab):

        #eps_mtx = self.map_eps_eng_to_mtx(eps_eng)
        MPTT_ijr = self._get__MPTT()
        return np.einsum('nijr,Emij->Emnr', MPTT_ijr, eps_Emab)

    def _get_e_equiv_Emn(self, e_Emna):
        '''
        Returns a list of the microplane equivalent strains
        based on the list of microplane strain vectors
        '''
        # magnitude of the normal strain vector for each microplane
        e_N_Emn = self._get_e_N_Emn(e_Emna)
        # print e_N_Emn[0, -1, :]
        # positive part of the normal strain magnitude for each microplane
        e_N_pos_Emn = (np.abs(e_N_Emn) + e_N_Emn) / 2.0
        # normal strain vector for each microplane
        e_N_Emna = np.einsum('Emn,ni -> Emni', e_N_Emn, self._MPN)
        # tangent strain ratio
        c_T = self.c_T
        # tangential strain vector for each microplane
        e_T_Emna = e_Emna - e_N_Emna
        # squared tangential strain vector for each microplane
        e_TT_Emn = np.einsum('Emni,Emni -> Emn', e_T_Emna, e_T_Emna)
        # print e_TT_Emn[0, -1, :]
        # equivalent strain for each microplane
        e_equiv_Emn = np.sqrt(e_N_pos_Emn * e_N_pos_Emn + c_T * e_TT_Emn)
        return e_equiv_Emn

    def update_state_variables(self, eps_Emab, kappa, omega):

        e_Emna = self._get_e_Emna(eps_Emab)
        eps_eq_Emn = self._get_e_equiv_Emn(e_Emna)
        f_trial_Emn = eps_eq_Emn - self.epsilon_0
        I = np.where(f_trial_Emn > 0)
        kappa[I] = eps_eq_Emn[I]
        omega[I] = self._get_omega(eps_eq_Emn[I])
        return I

    def _get_omega(self, kappa_Emn):
        '''
        Return new value of damage parameter
        @param kappa:
        '''
        omega_Emn = np.zeros_like(kappa_Emn)
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        I = np.where(kappa_Emn >= epsilon_0)
        omega_Emn[I] = (
            1.0 - (epsilon_0 / kappa_Emn[I] *
                   np.exp(-1.0 * (kappa_Emn[I] - epsilon_0) /
                          (epsilon_f - epsilon_0))
                   ))
        return omega_Emn

    def _get_phi_Emab(self, kappa_Emn):
        # Returns the 2nd order damage tensor 'phi_mtx'
        # scalar integrity factor for each microplane
        phi_Emn = 1.0 - self._get_omega(kappa_Emn)
        # integration terms for each microplanes
        phi_Emab = np.einsum(
            'Emn,n,nab->Emab',
            phi_Emn, self._MPW, self._MPNN
        )

        return phi_Emab

    #----------------------------------------------------------------
    #  the fourth order volumetric-np.identity( tensor
    #----------------------------------------------------------------
    def _get_I_vol_abcd(self):

        delta = np.identity(3)
        I_vol_abcd = (1.0 / 3.0) * np.einsum('ab,cd -> abcd', delta, delta)
        return I_vol_abcd

    #----------------------------------------------------------------
    # Returns the fourth order deviatoric-np.identity( tensor
    #----------------------------------------------------------------
    def _get_I_dev_abcd(self):

        delta = np.identity(3)
        I_dev_abcd = 0.5 * (np.einsum('ac,bd -> abcd', delta, delta) +
                            np.einsum('ad,bc -> abcd', delta, delta)) \
            - (1.0 / 3.0) * np.einsum('ab,cd -> abcd', delta, delta)

        return I_dev_abcd

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_vol_ab(self):

        delta = np.identity(3)
        P_vol_ab = (1.0 / 3.0) * delta
        return P_vol_ab

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_dev [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_dev_nabc(self):

        delta = np.identity(3)
        P_dev_nabc = 0.5 * \
            np.einsum('nd,da,bc -> nabc', self._MPN, delta, delta)
        return P_dev_nabc

    #----------------------------------------------------------------
    # Returns the outer product of P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_PP_vol_abcd(self):

        delta = np.identity(3)
        PP_vol_abcd = (1.0 / 9.0) * np.einsum('ab,cd -> abcd', delta, delta)
        return PP_vol_abcd

    #----------------------------------------------------------------
    # Returns the inner product of P_dev
    #----------------------------------------------------------------
    def _get_PP_dev_nabcd(self):
        delta = np.identity(3)
        PP_dev_nabcd = 0.5 * (
            0.5 * (np.einsum('na,nc,bd -> nabcd', self._MPN, self._MPN, delta) +
                   np.einsum('na,nd,bc -> nabcd', self._MPN, self._MPN, delta)) +
            0.5 * (np.einsum('ac,nb,nd -> nabcd',  delta, self._MPN, self._MPN) +
                   np.einsum('ad,nb,nc -> nabcd',  delta, self._MPN, self._MPN))) -\
            (1.0 / 3.0) * (np.einsum('na,nb,cd -> nabcd', self._MPN, self._MPN, delta) +
                           np.einsum('ab,nc,nd -> nabcd', delta, self._MPN, self._MPN)) +\
            (1.0 / 9.0) * np.einsum('ab,cd -> abcd', delta, delta)
        return PP_dev_nabcd

    def _get_I_vol_4(self):
        # The fourth order volumetric-np.identity( tensor
        delta = np.identity(3)
        I_vol_ijkl = (1.0 / 3.0) * np.einsum('ij,kl -> ijkl', delta, delta)
        return I_vol_ijkl

    def _get_I_dev_4(self):
        # The fourth order deviatoric-np.identity( tensor
        delta = np.identity(3)
        I_dev_ijkl = 0.5 * (np.einsum('ik,jl -> ijkl', delta, delta) +
                            np.einsum('il,jk -> ijkl', delta, delta)) \
            - (1 / 3.0) * np.einsum('ij,kl -> ijkl', delta, delta)

        return I_dev_ijkl

    def _get_P_vol(self):
        delta = np.identity(3)
        P_vol_ij = (1 / 3.0) * delta
        return P_vol_ij

    def _get_P_dev(self):
        delta = np.identity(3)
        P_dev_njkl = 0.5 * \
            np.einsum('ni,ij,kl -> njkl', self._MPN, delta, delta)
        return P_dev_njkl

    def _get_PP_vol_4(self):
        # outer product of P_vol
        delta = np.identity(3)
        PP_vol_ijkl = (1 / 9.) * np.einsum('ij,kl -> ijkl', delta, delta)
        return PP_vol_ijkl

    def _get_PP_dev_4(self):
        # inner product of P_dev
        delta = np.identity(3)
        PP_dev_nijkl = 0.5 * (
            0.5 * (np.einsum('ni,nk,jl -> nijkl', self._MPN, self._MPN, delta) +
                   np.einsum('ni,nl,jk -> nijkl', self._MPN, self._MPN, delta)) +
            0.5 * (np.einsum('ik,nj,nl -> nijkl',  delta, self._MPN, self._MPN) +
                   np.einsum('il,nj,nk -> nijkl',  delta, self._MPN, self._MPN))) -\
            (1 / 3.) * (np.einsum('ni,nj,kl -> nijkl', self._MPN, self._MPN, delta) +
                        np.einsum('ij,nk,nl -> nijkl', delta, self._MPN, self._MPN)) +\
            (1 / 9.) * np.einsum('ij,kl -> ijkl', delta, delta)
        return PP_dev_nijkl

    #--------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(29))
    #--------------------------------------------------------------------------
    def _get_S_1_Emabcd(self, eps_Emab, kappa, omega):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        phi_Emn = 1.0 - self._get_omega(kappa)
        # print 'phi_Emn', phi_Emn

        PP_vol_abcd = self._get_PP_vol_abcd()
        PP_dev_nabcd = self._get_PP_dev_nabcd()
        I_dev_abcd = self._get_I_dev_abcd()

#         PP_vol_abcd = self._get_PP_vol_4()
#         PP_dev_nabcd = self._get_PP_dev_4()
#         I_dev_abcd = self._get_I_dev_4()

        S_1_Emabcd = K0 * \
            np.einsum('Emn, n, abcd-> Emabcd', phi_Emn, self._MPW, PP_vol_abcd) + \
            G0 * 2.0 * self.zeta_G * np.einsum('Emn, n, nabcd-> Emabcd',
                                               phi_Emn, self._MPW, PP_dev_nabcd) - (1.0 / 3.0) * (
                2.0 * self.zeta_G - 1.0) * G0 * np.einsum('Emn, n, abcd-> Emabcd',
                                                          phi_Emn, self._MPW, I_dev_abcd)

        return S_1_Emabcd

    #-------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor with the (double orthotropic) assumption
    #-------------------------------------------------------------------------
    def _get_S_4_Emabcd(self, eps_Emab, kappa, omega):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)
        I_vol_abcd = self._get_I_vol_abcd()
        I_dev_abcd = self._get_I_dev_abcd()

        delta = np.identity(3)
        phi_Emab = self._get_phi_Emab(kappa)

        D_Emab = delta - phi_Emab

        d_Em = (1.0 / 3.0) * np.einsum('Emaa -> Em', D_Emab)

        D_bar_Emab = self.zeta_G * \
            (D_Emab - np.einsum('Em, ab -> Emab', d_Em, delta))

        S_4_Emabcd = K0 * I_vol_abcd - K0 * np.einsum('Em,abcd -> Emabcd',
                                                      d_Em, I_vol_abcd) +\
            G0 * I_dev_abcd - G0 * np.einsum('Em,abcd -> Emabcd',
                                             d_Em, I_dev_abcd) +\
            (2.0 / 3.0) * (G0 - K0) * (np.einsum('ij,Emkl -> Emijkl',
                                                 delta, D_bar_Emab) +
                                       np.einsum('Emij,kl -> Emijkl',
                                                 D_bar_Emab, delta)) + 0.5 * (- K0 + 2.0 * G0) *\
            (0.5 * (np.einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab) + np.einsum('Emil,jk -> Emijkl', D_bar_Emab, delta)) +
             0.5 * (np.einsum('Emil,jk -> Emijkl', D_bar_Emab, delta) + np.einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab)))

        return S_4_Emabcd

    #----------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor (restrctive orthotropic)
    #----------------------------------------------------------------------
    def _get_S_5_Emabcd(self, eps_Emab, kappa, omega):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        delta = np.identity(3)
        phi_Emab = self._get_phi_Emab(kappa)

        # damaged stiffness without simplification
        S_5_Emabcd = (1.0 / 3.0) * (K0 - G0) * 0.5 * ((np.einsum('ij,Emkl -> Emijkl', delta, phi_Emab) +
                                                       np.einsum('Emij,kl -> Emijkl', phi_Emab, delta))) + \
            G0 * 0.5 * ((0.5 * (np.einsum('ik,Emjl -> Emijkl', delta, phi_Emab) + np.einsum('Emil,jk -> Emijkl', phi_Emab, delta)) +
                         0.5 * (np.einsum('Emik,jl -> ijkl', phi_Emab, delta) + np.einsum('il,Emjk  -> Emijkl', delta, phi_Emab))))

        return S_5_Emabcd

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab, tn1, kappa, omega):

        I = self.update_state_variables(eps_Emab, kappa, omega)
        D_Emabcd = self._get_S_4_Emabcd(eps_Emab, kappa, omega)
        sig_Emab = np.einsum(
            'Emabcd,Emcd -> Emab',
            D_Emabcd, eps_Emab
        )

        return sig_Emab, D_Emabcd

    #-----------------------------------------------
    # number of microplanes - currently fixed for 3D
    #-----------------------------------------------
    n_mp = Constant(28)

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return np.array(
            [[.577350259, .577350259, .577350259],
             [.577350259, .577350259, -.577350259],
             [.577350259, -.577350259, .577350259],
             [.577350259, -.577350259, -.577350259],
             [.935113132, .250562787, .250562787],
             [.935113132, .250562787, -.250562787],
             [.935113132, -.250562787, .250562787],
             [.935113132, -.250562787, -.250562787],
             [.250562787, .935113132, .250562787],
             [.250562787, .935113132, -.250562787],
             [.250562787, -.935113132, .250562787],
             [.250562787, -.935113132, -.250562787],
             [.250562787, .250562787, .935113132],
             [.250562787, .250562787, -.935113132],
             [.250562787, -.250562787, .935113132],
             [.250562787, -.250562787, -.935113132],
             [.186156720, .694746614, .694746614],
             [.186156720, .694746614, -.694746614],
             [.186156720, -.694746614, .694746614],
             [.186156720, -.694746614, -.694746614],
             [.694746614, .186156720, .694746614],
             [.694746614, .186156720, -.694746614],
             [.694746614, -.186156720, .694746614],
             [.694746614, -.186156720, -.694746614],
             [.694746614, .694746614, .186156720],
             [.694746614, .694746614, -.186156720],
             [.694746614, -.694746614, .186156720],
             [.694746614, -.694746614, -.186156720]]
        )

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return np.array(
            [.0160714276, .0160714276, .0160714276, .0160714276, .0204744730,
             .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
             .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
             .0204744730, .0158350505, .0158350505, .0158350505, .0158350505,
             .0158350505, .0158350505, .0158350505, .0158350505, .0158350505,
             .0158350505, .0158350505, .0158350505]) * 6.0

    #-------------------------------------------------------------------------
    # Cached elasticity tensors
    #-------------------------------------------------------------------------




from traits.api import \
    HasTraits, List, Array, Property, cached_property, \
    Instance, Trait, Button, on_trait_change, \
    Int, Float, DelegatesTo, provides, WeakRef, Bool

from ibvpy.mesh.sdomain import \
    SDomain
# from ibvpy.view.plot3d.mayavi_util.pipelines import \
#     MVPolyData, MVPointLabels
import numpy as np

from .cell_array import CellView, CellArray, ICellArraySource
from .cell_grid import CellGrid
from .cell_grid_slice import CellGridSlice


#--------------------------------------------------------------------------
# DofGrid
#--------------------------------------------------------------------------
@provides(ICellArraySource)
class DofCellGrid(SDomain):

    '''
    Get an array with element Dof numbers
    '''
    cell_grid = Instance(CellGrid)

    get_cell_point_X_arr = DelegatesTo('cell_grid')
    get_cell_mvpoints = DelegatesTo('cell_grid')
    cell_node_map = DelegatesTo('cell_grid')
    get_cell_offset = DelegatesTo('cell_grid')

    # offset of dof within domain list
    #
    dof_offset = Int(0)

    # number of degrees of freedom in a single node
    #
    n_nodal_dofs = Int(3)
    #-------------------------------------------------------------------------
    # Generation methods for geometry and index maps
    #-------------------------------------------------------------------------
    n_dofs = Property(depends_on='cell_grid.shape,n_nodal_dofs,dof_offset')

    def _get_n_dofs(self):
        '''
        Get the total number of DOFs
        '''
        unique_cell_nodes = np.unique(self.cell_node_map.flatten())
        n_unique_nodes = len(unique_cell_nodes)
        return n_unique_nodes * self.n_nodal_dofs

    dofs = Property(depends_on='cell_grid.shape,n_nodal_dofs,dof_offset')

    @cached_property
    def _get_dofs(self):
        '''
        Construct the point grid underlying the mesh grid structure.
        '''
        cell_node_map = self.cell_node_map

        unique_cell_nodes = np.unique(cell_node_map.flatten())
        n_unique_nodes = len(unique_cell_nodes)

        n_nodal_dofs = self.n_nodal_dofs
        n_nodes = self.cell_grid.point_grid_size
        node_dof_array = np.repeat(-1, n_nodes *
                                   n_nodal_dofs).reshape(n_nodes, n_nodal_dofs)

        # Enumerate the DOFs in the mesh. The result is an array with n_nodes rows
        # and n_nodal_dofs columns
        #
        # A = array( [[ 0, 1 ],
        #             [ 2, 3 ],
        #             [ 4, 5 ]] );
        #
        node_dof_array[np.index_exp[unique_cell_nodes]] = \
            np.arange(
                n_unique_nodes * n_nodal_dofs).reshape(n_unique_nodes,
                                                       n_nodal_dofs)

        # add the dof_offset before returning the array
        #
        node_dof_array += self.dof_offset
        return node_dof_array

    dofs_Ia = Property()

    def _get_dofs_Ia(self):
        return self.dofs

    def _get_doffed_nodes(self):
        '''
        Get the indices of nodes containing DOFs. 
        '''
        cell_node_map = self.cell_node_map

        unique_cell_nodes = np.unique(cell_node_map.flatten())

        n_nodes = self.cell_grid.point_grid_size
        doffed_nodes = np.repeat(-1, n_nodes)

        doffed_nodes[np.index_exp[unique_cell_nodes]] = 1
        return np.where(doffed_nodes > 0)[0]

    #-----------------------------------------------------------------
    # Elementwise-representation of dofs
    #-----------------------------------------------------------------

    cell_dof_map = Property(depends_on='cell_grid.shape,n_nodal_dofs')

    def _get_cell_dof_map(self):
        return self.dofs[np.index_exp[self.cell_grid.cell_node_map]]

    dof_Eid = Property
    '''Mapping of Element, Node, Dimension -> DOF 
    '''

    def _get_dof_Eid(self):
        return self.cell_dof_map

    cell_grid_dof_map = Property(depends_on='cell_grid.shape,n_nodal_dofs')

    def _get_cell_grid_dof_map(self):
        return self.dofs[np.index_exp[self.cell_grid.cell_grid_node_map]]

    def get_cell_dofs(self, cell_idx):
        return self.cell_dof_map[cell_idx]

    elem_dof_map = Property(depends_on='cell_grid.shape,n_nodal_dofs')

    @cached_property
    def _get_elem_dof_map(self):
        el_dof_map = np.copy(self.cell_dof_map)
        tot_shape = el_dof_map.shape[0]
        n_entries = el_dof_map.shape[1] * el_dof_map.shape[2]
        elem_dof_map = el_dof_map.reshape(tot_shape, n_entries)
        return elem_dof_map

    def __getitem__(self, idx):
        '''High level access and slicing to the cells within the grid.

        The return value is a tuple with 
        1. array of cell indices
        2. array of nodes for each element
        3. array of coordinates for each node.
        '''
        dgs = DofGridSlice(dof_grid=self, grid_slice=idx)
        return dgs

    #-----------------------------------------------------------------
    # Spatial queries for dofs
    #-----------------------------------------------------------------

    def _get_dofs_for_nodes(self, nodes):
        '''Get the dof numbers and associated coordinates
        given the array of nodes.
        '''
        doffed_nodes = self._get_doffed_nodes()
#         print 'nodes'
#         print nodes
#         print 'doffed_nodes'
#         print doffed_nodes
        intersect_nodes = np.intersect1d(
            nodes, doffed_nodes, assume_unique=False)
        return (self.dofs[np.index_exp[intersect_nodes]],
                self.cell_grid.point_X_arr[np.index_exp[intersect_nodes]])

    def get_boundary_dofs(self):
        '''Get the boundary dofs and the associated coordinates
        '''
        nodes = [self.cell_grid.point_idx_grid[s]
                 for s in self.cell_grid.boundary_slices]
        dofs, coords = [], []
        for n in nodes:
            d, c = self._get_dofs_for_nodes(n)
            dofs.append(d)
            coords.append(c)
        return (np.vstack(dofs), np.vstack(coords))

    def get_all_dofs(self):
        nodes = self.cell_grid.point_idx_grid[...]
        return self._get_dofs_for_nodes(nodes)

    def get_left_dofs(self):
        nodes = self.cell_grid.point_idx_grid[0, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_right_dofs(self):
        nodes = self.cell_grid.point_idx_grid[-1, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_top_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, -1, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, 0, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_front_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, :, -1]
        return self._get_dofs_for_nodes(nodes)

    def get_back_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, :, 0]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_left_dofs(self):
        nodes = self.cell_grid.point_idx_grid[0, 0, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_front_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, 0, -1]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_back_dofs(self):
        nodes = self.cell_grid.point_idx_grid[:, 0, 0]
        return self._get_dofs_for_nodes(nodes)

    def get_top_left_dofs(self):
        nodes = self.cell_grid.point_idx_grid[0, -1, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_right_dofs(self):
        nodes = self.cell_grid.point_idx_grid[-1, 0, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_top_right_dofs(self):
        nodes = self.cell_grid.point_idx_grid[-1, -1, ...]
        return self._get_dofs_for_nodes(nodes)

    def get_bottom_middle_dofs(self):
        if self.cell_grid.point_idx_grid.shape[0] % 2 == 1:
            slice_middle_x = self.cell_grid.point_idx_grid.shape[0] / 2
            nodes = self.cell_grid.point_idx_grid[slice_middle_x, 0, ...]
            return self._get_dofs_for_nodes(nodes)
        else:
            print('Error in get_bottom_middle_dofs:'
                  ' the method is only defined for an odd number of dofs in x-direction')

    def get_top_middle_dofs(self):
        if self.cell_grid.point_idx_grid.shape[0] % 2 == 1:
            slice_middle_x = self.cell_grid.point_idx_grid.shape[0] / 2
            nodes = self.cell_grid.point_idx_grid[slice_middle_x, -1, ...]
            return self._get_dofs_for_nodes(nodes)
        else:
            print('Error in get_top_middle_dofs:'
                  ' the method is only defined for an odd number of dofs in x-direction')

    def get_left_middle_dofs(self):
        if self.cell_grid.point_idx_grid.shape[1] % 2 == 1:
            slice_middle_y = self.cell_grid.point_idx_grid.shape[1] / 2
            nodes = self.cell_grid.point_idx_grid[0, slice_middle_y, ...]
            return self._get_dofs_for_nodes(nodes)
        else:
            print('Error in get_left_middle_dofs:'
                  ' the method is only defined for an odd number of dofs in y-direction')

    def get_right_middle_dofs(self):
        if self.cell_grid.point_idx_grid.shape[1] % 2 == 1:
            slice_middle_y = self.cell_grid.point_idx_grid.shape[1] / 2
            nodes = self.cell_grid.point_idx_grid[-1, slice_middle_y, ...]
            return self._get_dofs_for_nodes(nodes)
        else:
            print('Error in get_right_middle_dofs:'
                  ' the method is only defined for an odd number of dofs in y-direction')

    def get_left_front_bottom_dof(self):
        nodes = self.cell_grid.point_idx_grid[0, 0, -1]
        return self._get_dofs_for_nodes(nodes)

    def get_left_front_middle_dof(self):
        if self.cell_grid.point_idx_grid.shape[1] % 2 == 1:
            slice_middle_y = self.cell_grid.point_idx_grid.shape[1] / 2
            nodes = self.cell_grid.point_idx_grid[0, slice_middle_y, -1]
            return self._get_dofs_for_nodes(nodes)
        else:
            print('Error in get_left_middle_front_dof:'
                  ' the method is only defined for an odd number of dofs in y-direction')

    #-----------------------------------------------------------------
    # Visualization related methods
    #-----------------------------------------------------------------

    refresh_button = Button('Draw')

    @on_trait_change('refresh_button')
    def redraw(self):
        '''Redraw the point grid.
        '''
        self.cell_grid.redraw()

    dof_cell_array = Button

    def _dof_cell_array_fired(self):
        cell_array = self.cell_grid.cell_node_map
        self.show_array = CellArray(data=cell_array,
                                    cell_view=DofCellView(cell_grid=self))
        self.show_array.current_row = 0
        self.show_array.configure_traits(kind='live')

class DofGridSlice(CellGridSlice):

    dof_grid = WeakRef(DofCellGrid)

    def __init__(self, dof_grid, **args):
        self.dof_grid = dof_grid
        super(DofGridSlice, self).__init__(**args)

    cell_grid = Property()

    def _get_cell_grid(self):
        return self.dof_grid.cell_grid

    dofs = Property

    def _get_dofs(self):
        _, idx2 = self.idx_tuple
        return self.dof_grid.cell_dof_map[
            np.ix_(
                self.elems,
                self.cell_grid.grid_cell[idx2]
            )
        ]

#-----------------------------------------------------------------------
# View a single cell instance
#-----------------------------------------------------------------------


class DofCellView(CellView):

    '''View a single cell instance.
    '''
    # implements(ICellView)

    elem_dofs = Array

    def set_cell_traits(self):
        '''Set the trait values for the current cell_idx
        '''
        self.elem_dofs = self.cell_grid.get_cell_dofs(self.cell_idx)


    def _get_cell_mvpoints(self):
        return self.cell_grid.get_cell_mvpoints(self.cell_idx)

    def _get_cell_labels(self):
        cell_dofs = self.cell_grid.get_cell_dofs(self.cell_idx)
        shape = cell_dofs.shape
        if shape[1] < 3:
            cd = np.zeros((shape[0], 3))
            cd[:, :shape[1]] = cell_dofs
            return cd
        else:
            return cell_dofs

    def redraw(self):
        if self.draw_cell:
            self.mvp_elem_labels.redraw(label_mode='label_vectors')




#-------------------------------------------------------------------------------
#
# Copyright (c) 2020, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.
#
# Thanks for using BMCS open source!
#

from ibvpy.fets import FETSEval

class FETS1D5(FETSEval):
    '''Base class for 2D elements.
    '''




'''
Created on Sep 3, 2009

'''

import copy
from math import pi

from ibvpy.tmodel.mats3D.mats3D_tensor import \
    map3d_eps_eng_to_mtx, map3d_sig_eng_to_mtx, map3d_eps_mtx_to_eng, map3d_sig_mtx_to_eng, \
    map3d_ijkl2mn, map3d_tns2_to_tns4, map3d_tns4_to_tns2, compliance_mapping3d
from traits.api import Callable, Constant, Property

from ibvpy.tmodel.matsXD.vmatsXD_eval import MATSXDEval


class MATS3DEval(MATSXDEval):
    '''Base class for 3D models
    '''

    # number of spatial dimensions of an integration cell for the material model
    #
    n_dims = Constant(3)


class NotUsedAnyMore:
    # dimension dependent tensor mappings
    #
    map_tns4_to_tns2 = Callable(map3d_tns4_to_tns2, transient=True)
    map_eps_eng_to_mtx = Callable(map3d_eps_eng_to_mtx, transient=True)
    map_sig_eng_to_mtx = Callable(map3d_sig_eng_to_mtx, transient=True)
    compliance_mapping = Callable(compliance_mapping3d, transient=True)
    map_sig_mtx_to_eng = Callable(map3d_sig_mtx_to_eng, transient=True)
    map_eps_mtx_to_eng = Callable(map3d_eps_mtx_to_eng, transient=True)

    def _get_explorer_config(self):
        '''Get the specific configuration of this material model in the explorer
        '''
        c = super(MATS3DEval, self)._get_explorer_config()

        from ibvpy.api import TLine
        from ibvpy.tmodel.mats3D.mats3D_explorer_bcond import BCDofProportional3D

        # overload the default configuration
        c['bcond_list'] = [
            BCDofProportional3D(max_strain=0.005, phi=0., theta=pi / 2.)]
        c['tline'] = TLine(step=0.05, max=1)
        return c



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch


from traits.api import \
    HasTraits, Instance, List

import unittest

from ibvpy.fets.fets2D5.fets2D58h import \
    FETS2D58H
from ibvpy.util.simgrid import \
    simgrid
from ibvpy.tmodel.mats_eval import \
    MATSEval
from numpy import \
    array

class TestMATS2D5(unittest.TestCase, HasTraits):

    mats_eval = Instance( MATSEval )
    
    # values of displacements that must equal for the symmetry tests
    sym_assert_pattern = List( [( (0,0), (1,1) ),
                                ( (0,1), (1,0) ),
                                ( (0,2), (1,2) )] )
    
    def assert_2D_symmetry_clamped_cube(self, load_dirs, load = 0.001 ):
        '''Assert that the symmetry is given for the applied loadings.
        '''

        self.fets_eval  = FETS2D58H(mats_eval  = self.mats_eval)

        support_slices = [
                      [ 
                       (0   ,slice(None),slice(None),0   ,slice(None),slice(None)), #                             yz plane  0
                      ],
                      [ 
                       (slice(None),0   ,slice(None),slice(None),0   ,slice(None)), # xz plane  1
                      ],
                      ]
        support_dirs = [[0,1,2]]
        
        loading_slices = [
                      (-1, slice(None),slice(None),-1 ,slice(None),slice(None)),  # loading in x dir
                      (slice(None),-1 ,slice(None),slice(None),-1 ,slice(None)),  # loading in y dir
                        ]

        vars = [] 

        r = [ simgrid( self.fets_eval,  (1,1,1), (1,1,1),
                         support_slice, support_dirs,
                         loading_slice, load_dir, 
                         load, 1, vars )
                for support_slice, loading_slice, load_dir in zip( support_slices,
                                                                   loading_slices, 
                                                                   load_dirs ) ]

        u = array( [ r[i][1][-3:] for i in range(2) ] )
        for idx1, idx2 in self.sym_assert_pattern: 
            self.assertAlmostEqual( u[idx1], u[idx2] )

    def assert_stress_value(self,
                            sig_expected,  
                            n_steps = 3, 
                            load = 0.0001,
                            ):
        '''Assert that the symmetry is given for the applied loadings.
        '''
        self.fets_eval  = FETS2D58H(mats_eval  = self.mats_eval)        
        support_slices = [
                      [ 
                       (0   ,slice(None),slice(None),0   ,slice(None),slice(None)), # yz plane  0
                       (0   ,0          ,slice(None),0   ,0          ,slice(None)), # z plane  0
                       (0   ,0          ,0          ,0   ,0          ,0          ), # z plane  0
                      ],
                      ]
        support_dirs = [[0],[1],[2]]
        
        loading_slices = [
                      (-1, slice(None),slice(None),-1 ,slice(None),slice(None)),  # loading in x dir
                        ]

        load_dirs = [0]

        vars = [ 'sig_app' ]

        r = [ simgrid( self.fets_eval,  (1,1,1), (1,1,1),
                       support_slice, support_dirs,
                       loading_slice, load_dir, 
                       load, 1, vars )
                for support_slice, loading_slice, load_dir in zip( support_slices,
                                                                   loading_slices, 
                                                                   load_dirs ) ]

        for rr in r:
            sig_end = rr[2][0]
            # all material points must have the same value - uniform loading.
            for sig, sig_exp in zip( sig_end.flatten(), sig_expected ):
                self.assertAlmostEqual( sig, sig_exp, 5 )




from numpy import array, ones, trapz, frompyfunc, dot

from traits.api import List, Callable
from ibvpy.api import RTDofGraph


class MATS2DMicroplaneDamageTraceGfmac(RTDofGraph):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''

    _trace = List()
    var_time_eval = Callable

    def __init__(self, **kwrds):
        super(MATS2DMicroplaneDamageTraceGfmac, self).__init__(**kwrds)
        self._trace = []

    def bind(self):
        '''
        Locate the evaluators
        '''
        super(MATS2DMicroplaneDamageTraceGfmac, self).bind()
        # get the evaluator for the variable
        self.var_time_eval = self.rmgr.rte_dict['time']

    def add_current_values(self, sctx, eps_eng):
        '''
        Invoke the evaluators in the current context for the specified control vector U_k.
        '''
        super(MATS2DMicroplaneDamageTraceGfmac, self).add_current_values(
            sctx, eps_eng)
        time_value = self.var_time_eval(sctx, eps_eng)
        self._trace.append(time_value[0])

    def redraw(self):
        if self.idx_x < 0 or self.idx_y < 0 or \
                self._xdata == [] or self._ydata == []:
            return

        xarray = array(self._xdata)[:, self.idx_x]
        yarray = array(self._ydata)[:, self.idx_y]
        print('in Gf_mac: ydata', yarray)

        # macroscopic total energy:
        E_t = self._get_E_t(xarray, yarray)
        print('E_t_mac = %.10f' % (E_t))

        # macroscopic elastic energy:
        U_t = self._get_U_t(xarray, yarray)
        print('U_t_mac = %.10f' % (U_t))

        G_f = E_t - U_t
        print('G_f_mac = %.10f' % (G_f))

        self.trace.xdata = array(self._trace, dtype=float)
        self.trace.ydata = G_f * ones(self.trace.xdata.shape, dtype=float)

    def _get_E_t(self, xdata, ydata):
        # integral under the stress strain curve
        E_t = trapz(ydata, xdata)
        return E_t

    def _get_U_t(self, xdata, ydata):
        # area of the stored elastic energy
        U_t = 0.0
        if len(xdata) != 0:
            U_t = 0.5 * ydata[-1] * xdata[-1]
        return U_t

    def clear(self):
        super(MATS2DMicroplaneDamageTraceGfmac, self).clear()
        self._trace = []


class MATS2DMicroplaneDamageTraceEtmac(MATS2DMicroplaneDamageTraceGfmac):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''

    def redraw(self):
        if self.idx_x < 0 or self.idx_y < 0 or \
                self._xdata == [] or self._ydata == []:
            return
        #
        xarray = array(self._xdata)[:, self.idx_x]
        yarray = array(self._ydata)[:, self.idx_y]
        # macroscopic total energy:
        E_t = self._get_E_t(xarray, yarray)
        #
        self.trace.xdata = array(self._trace, dtype=float)
        self.trace.ydata = E_t * ones(self.trace.xdata.shape, dtype=float)


class MATS2DMicroplaneDamageTraceUtmac(MATS2DMicroplaneDamageTraceGfmac):
    '''Evaluate the energy contributions spread over 
    several fields of stress-strain responses. 
    '''

    def redraw(self):
        if self.idx_x < 0 or self.idx_y < 0 or \
                self._xdata == [] or self._ydata == []:
            return
        #
        xarray = array(self._xdata)[:, self.idx_x]
        yarray = array(self._ydata)[:, self.idx_y]
        # macroscopic total energy:
        U_t = self._get_U_t(xarray, yarray)
        #
        self.trace.xdata = array(self._trace, dtype=float)
        self.trace.ydata = U_t * ones(self.trace.xdata.shape, dtype=float)



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jan 20, 2011 by: rch

from numpy import \
    array, tensordot, dot, zeros, c_, ix_, mgrid, arange, \
    where, sum, sin, cos, vstack, hstack, argmax, newaxis, size, \
    shape, sqrt, frompyfunc, ones_like, zeros_like, ones, any, all, \
    sort, argsort, concatenate, add
from numpy import where, zeros_like, fabs, sign, ma
from traits.api import \
    HasTraits, Float, Array, Property, cached_property, Instance, Enum, \
    Dict, Bool, Int, Callable


class GeoSquare2Circle(HasTraits):
    '''Geometry definition.
    '''
    circle_center = Array('float64', value=[0.0, 0.0])
    circle_radius = Float(0.38)
    square_edge = Float(2.0)

    pre_transform = Callable

    post_transform = Callable

    def __call__(self, points):
        if self.pre_transform:
            points = self.pre_transform(points)

        r, s = self._get_axes_data(points)
        r_t, s_t = self._transform(r, s)
        self._set_axes_data(points, r_t, s_t)

        if self.post_transform:
            points = self.post_transform(points)

        return points

    def _get_axes_data(self, points):
        return points[:, 0], points[:, 1]

    def _set_axes_data(self, points, r, s):
        points[:, 0] = r
        points[:, 1] = s

    def _transform(self, r, s):
        '''Return the global coordinates of the supplied local points.
        '''

        R = self.circle_radius
        D = self.square_edge / 2.0

        cx, cy = self.circle_center

        # number of local grid points for each coordinate direction
        # values must range between 0 and 1
        Xi, Yi = r - cx, s - cy

        # delimit the area within the square edge
        mask = where((fabs(Xi) <= D) & (fabs(Yi) <= D))

        xi, yi = Xi[ mask ], Yi[ mask ]
        X, Y = zeros_like(xi), zeros_like(yi)

        idx_center = where(xi ** 2 + yi ** 2 == 0)

        # size of total structure
        #
        idx_Rx = where(fabs(xi) > fabs(yi))
        idx_Ry = where(fabs(yi) >= fabs(xi))

        Px, Py = xi[ idx_Rx ], yi[ idx_Rx ]
        SPxy2 = sqrt(Px ** 2 + Py ** 2)

        X[idx_Rx] = sign(Px) * Px ** 2 / SPxy2
        Y[idx_Rx] = sign(Py) * fabs(Px) * fabs(Py) / SPxy2

        Px, Py = yi[ idx_Ry ], xi[ idx_Ry ]
        SPxy2 = sqrt(Px ** 2 + Py ** 2)

        Y[idx_Ry] = sign(Px) * Px ** 2 / SPxy2
        X[idx_Ry] = sign(Py) * fabs(Px) * fabs(Py) / SPxy2

        X[ idx_center ] = 0.0
        Y[ idx_center ] = 0.0

        delta_x, delta_y = X - xi, Y - yi

        # perform the transition mapping from the circle to the square

        idx_corner = where((fabs(xi) >= R) & (fabs(yi) >= R))
        idx_right = where((fabs(yi) < R) & (fabs(xi) >= R))
        idx_top = where((fabs(xi) < R) & (fabs(yi) >= R))

        xii, yii = xi[ idx_corner], yi[idx_corner ]
        print(R)
        print(D)
        a = 1.0 / (R ** 2 - 2 * D * R + D ** 2)
        b = -D / (R ** 2 - 2 * D * R + D ** 2)
        c = -D / (R ** 2 - 2 * D * R + D ** 2)
        d = D ** 2 / (R ** 2 - 2 * D * R + D ** 2)
        weight = (a * fabs(xii) * fabs(yii) + b * fabs(xii) + c * fabs(yii) + d)
        dx, dy = delta_x[ idx_corner ], delta_y[idx_corner]
        X[ idx_corner ], Y[ idx_corner ] = xii + weight * dx, yii + weight * dy

        xii, yii = xi[ idx_right], yi[idx_right ]
        weight = 1.0 / (R - D) * (fabs(xii) - D)
        dx, dy = delta_x[ idx_right ], delta_y[idx_right]
        X[ idx_right ], Y[ idx_right ] = xii + weight * dx, yii + weight * dy

        xii, yii = xi[ idx_top], yi[idx_top ]
        weight = 1.0 / (R - D) * (fabs(yii) - D)
        dx, dy = delta_x[ idx_top ], delta_y[idx_top]
        X[ idx_top ], Y[ idx_top ] = xii + weight * dx, yii + weight * dy

        Xi[mask] = X
        Yi[mask] = Y

        return Xi + cx, Yi + cy


if __name__ == '__main__':

    s2c = GeoSquare2Circle(circle_center=[0.2, 0.2], circle_radius=0.4, square_edge=2.0)
    print('result', s2c(array([[ 0.3, 0.5, 0 ]], dtype=float)))




from functools import reduce

from ibvpy.fets.i_fets import IFETSEval
from ibvpy.mesh.cell_grid.cell_array import ICellView, CellView, CellArray, ICellArraySource
from ibvpy.mesh.cell_grid.cell_grid import CellGrid
from ibvpy.mesh.cell_grid.cell_spec import CellSpec
from ibvpy.mesh.cell_grid.dof_grid import DofCellGrid, DofCellView
from ibvpy.mesh.cell_grid.geo_grid import GeoCellGrid, GeoCellView
from numpy import copy, zeros, array_equal
from traits.api import \
    Instance, Array, Int, on_trait_change, Property, cached_property, \
    List, Button, HasTraits, provides, WeakRef, Float,  \
    Callable, Str, Event
from .fe_grid_activation_map import FEGridActivationMap
from .fe_grid_idx_slice import FEGridIdxSlice
from .fe_grid_ls_slice import FEGridLevelSetSlice
from .fe_grid_node_slice import ISliceProperty
from .i_fe_uniform_domain import IFEUniformDomain



class MElem(HasTraits):

    '''
    Finite element spatial representation.
    '''
    point_X_arr = Array
    point_x_arr = Array
    dofs = Array

    def get_X_mtx(self):
        '''
        Index mapping from the global array of coordinates.
        '''
        return self.point_X_arr

    def get_x_mtx(self):
        '''
        Index mapping from the global array of coordinates.
        '''
        return self.point_x_arr

    def get_dof_map(self):
        '''
        Return the dof map for the current element as a list
        '''
        return self.dofs

    def __str__(self):
        return 'points:\n%s\ndofs %s' % (self.point_X_arr, self.dofs)


@provides(ICellArraySource, IFEUniformDomain, ICellView)
class FEGrid(FEGridActivationMap):

    '''Structured FEGrid consisting of potentially independent
    dof_grid and geo_grid.

    For isoparametric element formulations, the dof_grid and geo_grid may
    share a single cell_grid to save memory.

    Structure of the grid
    ---------------------
    The structure of the grid is defined at two levels:
    1) within a cell specify the distribution of points
    (for dof_r and for geo_r).

    2) the cells are repeated in respective dimension by n_elem
    number of elements

    Services
    --------
    1) For a given element number return the nodal coordinates respecting
       the specification in the geo_r
    2) For a given element number return the array of dof numbers respecting
       the specification in the geo_r
    3) For a given CellSpec return a CellGrid respecting the geometric
       parameters of the FEGrid (applicable for response trace fields
       with finer distribution of nodes.
    4) For a given subdivision of the cell return CellGrid usable as
       visualization field.(where to specify the topology? - probably in
       the CellSpec?
    '''

    # Grid geometry specification
    #
    coord_min = Array(Float, value=[0., 0., 0.])
    coord_max = Array(Float, value=[1., 1., 1.])

    geo_transform = Callable

    # number of elements in the individual dimensions
    shape = Array(Int, value=[1, 1, 1], changes_ndofs=True)

    changed_structure = Event

    @on_trait_change('+changed_structure')
    def set_changed_structure(self):
        self.changed_structure = True

    _tree_label = 'subgrid'

    fets_eval = Instance(IFETSEval)

    # identifier within the refinement level
    idx = Int()

    # links within the dependency
    prev_grid = WeakRef('ibvpy.mesh.fe_grid.FEGrid', allow_none=True)
    next_grid = WeakRef('ibvpy.mesh.fe_grid.FEGrid', allow_none=True)

    _name = Str('')
    name = Property

    def _get_name(self):
        '''Return the name within the level
        '''
        if self._name == '':
            return 'grid ' + str(self.idx)
        else:
            return self._name

    def _set_name(self, value):
        self._name = value

    def __repr__(self):
        return self.name

    # dof offset within the global enumeration
    dof_offset = Property(
        Int, depends_on='prev_grid.dof_offset,level.dof_offset')
    # cached_property

    def _get_dof_offset(self):
        if self.prev_grid:
            return self.prev_grid.dof_offset + self.prev_grid.n_dofs
        elif self.level:
            return self.level.dof_offset
        else:
            return 0

    _level = WeakRef(
        'ibvpy.mesh.fe_refinement_grid.FERefinementGrid', links_changed=True)
    # parent domain
    level = Property

    def _set_level(self, value):
        'reset the parent of this domain'
        if self._level and self._level != value:
            # remove the grid from the level
            self._level._fe_subgrids.remove(self)
            # unlink it from previous and next
            if self.prev_grid:
                self.prev_grid.next_grid = self.next_grid
            if self.next_grid:
                self.next_grid.prev_grid = self.prev_grid
        # set the new parent
        self._level = value
        # link the subgrid within the level
#        prev_grid = self._level.last_subgrid
#        if prev_grid:
#            self.prev_grid = prev_grid
#            prev_grid.next_grid = self
        # add to the level
        self.idx = len(self._level._fe_subgrids)
        self._level._fe_subgrids.append(self)

    def _get_level(self):
        return self._level

    def __del__(self):
        ''' Release the grid from the dependency structure
        '''
        if self.prev_grid:
            self.prev_grid.next_grid = self.next_grid
        if self.next_grid:
            self.next_grid.prev_grid = self.prev_grid

    dots = Property

    @cached_property
    def _get_dots(self):
        '''Construct and return a new instance of domain
        time stepper.
        '''
        return self.fets_eval.dots_class(sdomain=self)

    dof_r = Property

    def _get_dof_r(self):
        return self.fets_eval.dof_r

    geo_r = Property

    def _get_geo_r(self):
        return self.fets_eval.geo_r

    n_nodal_dofs = Property

    def _get_n_nodal_dofs(self):
        return self.fets_eval.n_nodal_dofs

    #-------------------------------------------------------------------------
    # Derived properties
    #-------------------------------------------------------------------------
    # dof point distribution within the cell converted into the CellSpec format
    # CellSpec can derive the shape of the single grid cell, i.e.
    # the number of points specified in the individual directions.
    #
    dof_grid_spec = Property(Instance(CellSpec), depends_on='fets_eval.dof_r')

    def _get_dof_grid_spec(self):
        return CellSpec(node_coords=self.dof_r)

    # geo point distribution within the cell ... the same as above ...
    #
    geo_grid_spec = Property(Instance(CellSpec), depends_on='fets_eval.geo_r')

    def _get_geo_grid_spec(self):
        return CellSpec(node_coords=self.geo_r)

    # dof_grid is represented by the DofCellGrid class maintaining
    # the internal data structure to retrieve the element mappings
    # within the grid.
    #
    dof_grid = Property(Instance(DofCellGrid),
                        depends_on='fets_eval.dof_r,shape,coord_min,coord_max')

    def _get_dof_grid(self):
        return self._grids[0]

    dof_vertex_X_grid = Property

    def _get_dof_vertex_X_grid(self):
        return self.dof_grid.cell_grid.vertex_X_grid

    intg_elem_grid = Property

    def _get_intg_elem_grid(self):
        return self.dof_grid.cell_grid.cell_idx_grid

    ls_mask = Property

    def _get_ls_mask(self):
        return zeros(self.intg_elem_grid.shape, dtype=bool)

    ls_elem_grid = Property

    def _get_ls_elem_grid(self):
        return self.dof_grid.cell_grid.cell_idx_grid

    intg_elem_grid_dof_map = Property

    def _get_intg_elem_grid_dof_map(self):
        return self.dof_grid.cell_grid_dof_map

    def get_cell_point_X_arr(self, elem):
        return self.geo_grid.get_cell_point_X_arr(elem)

    # geo_grid is represented by the GeoCellGrid class maintaining
    # the internal data structure to retrieve the element mappings
    # within the grid.
    #
    geo_grid = Property(Instance(GeoCellGrid),
                        depends_on='fets_eval.geo_r,shape,coord_min,coord_max')

    def _get_geo_grid(self):
        return self._grids[1]

    I = Property

    def _get_I(self):
        return ISliceProperty(fe_grid=self)

    def __getitem__(self, idx):
        if isinstance(idx, tuple) or isinstance(idx, int):
            return FEGridIdxSlice(fe_grid=self, grid_slice=idx)
        elif isinstance(idx, str):
            return FEGridLevelSetSlice(fe_grid=self, ls_function_eval=idx)
        else:
            raise TypeError(type(idx), 'is unsupported type for slicing')

    #-------------------------------------------------------------
    # Implement the FEDomain interface
    #-------------------------------------------------------------

    n_dofs = Property(Int)

    def _get_n_dofs(self):
        return self.dof_grid.n_dofs

    def get_cell_offset(self, idx_tuple):
        return self.dof_grid.get_cell_offset(idx_tuple)
    #-------------------------------------------------------------------------
    # Implement the parent interface
    #-------------------------------------------------------------------------

    def deactivate(self, idx):
        '''Exclude the specified element from the integration.
        '''
        if isinstance(idx, tuple):
            self.inactive_elems.append(self.dof_grid.get_cell_offset(idx))
        elif isinstance(idx, int):
            self.inactive_elems.append(idx)
        if self.level:
            self.level.set_changed_structure()

    #-------------------------------------------------------------
    # Implement the IFEUniformDomain interface
    #-------------------------------------------------------------

    # Full elem dof map ignoring masked elements.
    ls_elem_dof_map = Property

    def _get_ls_elem_dof_map(self):
        return self.dof_grid.elem_dof_map

    # Full elem dof map ignoring masked elements.
    elem_dof_map_unmasked = Property

    def _get_elem_dof_map_unmasked(self):
        return self.dof_grid.elem_dof_map

    elem_dof_map = Property(
        depends_on='fets_eval.dof_r,shape,dof_offset, changed_structure')

    def _get_elem_dof_map(self):
        elem_dof_map = self.dof_grid.elem_dof_map[self.activation_map, :]
        return elem_dof_map

    elem_X_map = Property(
        depends_on='fets_eval.geo_r,shape,coord_min,coord_max,n_nodal_dofs, changed_structure')

    @cached_property
    def _get_elem_X_map(self):
        elem_X_map = self.geo_grid.elem_X_map[self.activation_map, :].copy()
        return elem_X_map

    ls_elem_X_map = Property

    def _get_ls_elem_X_map(self):
        return self.geo_grid.elem_X_map.copy()

    elem_X_map_unmasked = Property

    def _get_elem_X_map_unmasked(self):
        return self.geo_grid.elem_X_map.copy()

    elem_x_map = Property(
        depends_on='fets_eval.geo_r,shape,coord_min,coord_max,n_nodal_dofs, changed_structure')

    @cached_property
    def _get_elem_x_map(self):
        elem_x_map = self.geo_grid.elem_x_map[self.activation_map, :].copy()
        return copy(elem_x_map)

    elem_x_map_unmasked = Property

    def _get_elem_x_map_unmasked(self):
        return self.geo_grid.elem_x_map.copy()

    ls_elem_x_map = Property

    def _get_ls_elem_x_map(self):
        return self.geo_grid.elem_x_map.copy()

    X_dof_arr = Property

    def _get_X_dof_arr(self):
        return self.dof_grid.cell_grid.point_X_arr

    elem_node_map = Property

    def _get_elem_node_map(self):
        return self.dof_grid.cell_grid.cell_node_map

    n_grid_elems = Property

    def _get_n_grid_elems(self):
        return reduce(lambda x, y: x * y, self.shape)

    n_active_elems = Property

    def _get_n_active_elems(self):
        return self.elem_dof_map.shape[0]

    elements = Property(
        List, depends_on='fets_eval.dof_r,fets_eval.geo_r,shape+,coord_min,coord_max,fets_eval.n_nodal_dofs,dof_offset, changed_structure')

    @cached_property
    def _get_elements(self):
        return [MElem(dofs=dofs,
                      point_X_arr=point_X_arr,
                      point_x_arr=point_x_arr)
                for dofs, point_X_arr, point_x_arr
                in zip(self.elem_dof_map,
                       self.elem_X_map,
                       self.elem_x_map)]

    dof_Eid = Property
    '''Mapping of Element, Node, Dimension -> DOF 
    '''

    def _get_dof_Eid(self):
        return self.dof_grid.dof_Eid

    dofs = Property

    def _get_dofs(self):
        return self.dof_grid.dofs

    I_Ei = Property
    '''For a given element and its node number return the global index
    of the node'''

    def _get_I_Ei(self):
        return self.geo_grid.cell_grid.cell_node_map

    X_Id = Property()
    '''Array of containing the coordinate d \in (0,1,3)
    of a node I
    '''

    def _get_X_Id(self):
        return self.geo_grid.point_X_arr

    def apply_on_ip_grid(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element.
        @param ip_mask: specifies the local coordinates within the element.
        '''
        X_el = self.elem_X_map
        # test call to the function with single output - to get the shape of
        # the result.
        out_single = fn(ip_mask[0], X_el[0])
        out_grid_shape = (X_el.shape[0], ip_mask.shape[0],) + out_single.shape
        out_grid = zeros(out_grid_shape)

        for el in range(X_el.shape[0]):
            for ip in range(ip_mask.shape[0]):
                out_grid[el, ip, ...] = fn(ip_mask[ip], X_el[el])

        return out_grid

    def apply_on_ip_grid_unmasked(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element.
        @param ip_mask: specifies the local coordinates within the element.
        '''
        X_el = self.geo_grid.elem_X_map
        # test call to the function with single output - to get the shape of
        # the result.
        out_single = fn(ip_mask[0], X_el[0])
        out_grid_shape = (X_el.shape[0], ip_mask.shape[0],) + out_single.shape
        out_grid = zeros(out_grid_shape)

        for el in range(X_el.shape[0]):
            for ip in range(ip_mask.shape[0]):
                out_grid[el, ip, ...] = fn(ip_mask[ip], X_el[el])

        return out_grid

    #-------------------------------------------------------------
    # Queries
    #-------------------------------------------------------------
    # The delegation had to be replaced with explicit calls to functions.
    # The reason is lost binding if a method
    #
    # fe_domain.get_left_dofs
    #
    # If the discretization parameter changed, then a new dof_grid was generated
    # but the call fe_domain.get_left_dofs went to the original dof_grid.
    #
    # get_all_dofs   = DelegatesTo('dof_grid')
    def get_all_dofs(self): return self.dof_grid.get_all_dofs()
    # get_left_dofs   = Delegate('dof_grid')

    def get_left_dofs(self): return self.dof_grid.get_left_dofs()
    # get_right_dofs  = Delegate('dof_grid')

    def get_right_dofs(self): return self.dof_grid.get_right_dofs()
    # get_top_dofs    = Delegate('dof_grid')

    def get_top_dofs(self): return self.dof_grid.get_top_dofs()
    # get_bottom_dofs = Delegate('dof_grid')

    def get_bottom_dofs(self): return self.dof_grid.get_bottom_dofs()
    # get_front_dofs  = Delegate('dof_grid')

    def get_front_dofs(self): return self.dof_grid.get_front_dofs()
    # get_back_dofs   = Delegate('dof_grid')

    def get_back_dofs(self): return self.dof_grid.get_back_dofs()
    # get_bottom_left_dofs   = Delegate('dof_grid')

    def get_bottom_left_dofs(self): return self.dof_grid.get_bottom_left_dofs()
    # get_bottom_front_dofs   = Delegate('dof_grid')

    def get_bottom_front_dofs(
        self): return self.dof_grid.get_bottom_front_dofs()
    # get_bottom_back_dofs   = Delegate('dof_grid')

    def get_bottom_back_dofs(self): return self.dof_grid.get_bottom_back_dofs()
    # get_top_left_dofs   = Delegate('dof_grid')

    def get_top_left_dofs(self): return self.dof_grid.get_top_left_dofs()
    # get_bottom_right_dofs  = Delegate('dof_grid')

    def get_bottom_right_dofs(
        self): return self.dof_grid.get_bottom_right_dofs()
    # get_top_right_dofs  = Delegate('dof_grid')

    def get_top_right_dofs(self): return self.dof_grid.get_top_right_dofs()
    # get_bottom_middle_dofs  = Delegate('dof_grid')

    def get_bottom_middle_dofs(
        self): return self.dof_grid.get_bottom_middle_dofs()
    # get_top_middle_dofs  = Delegate('dof_grid')

    def get_top_middle_dofs(self): return self.dof_grid.get_top_middle_dofs()
    # get_left_middle_dofs  = Delegate('dof_grid')

    def get_left_middle_dofs(self): return self.dof_grid.get_left_middle_dofs()
    # get_right_middle_dofs  = Delegate('dof_grid')

    def get_right_middle_dofs(
        self): return self.dof_grid.get_right_middle_dofs()
    # get_left_front_bottom_dof  = Delegate('dof_grid')

    def get_left_front_bottom_dofs(
        self): return self.dof_grid.get_left_front_bottom_dofs()
    # get_left_front_middle_dof  = Delegate('dof_grid')

    def get_left_front_middle_dofs(
        self): return self.dof_grid.get_left_front_middle_dofs()

    #-------------------------------------------------------------
    # LevelSet Methods
    #-------------------------------------------------------------

    def get_lset_subdomain(self, lset_function):
        '''@TODO - implement the subdomain selection method
        '''
        raise NotImplementedError

    def get_boundary(self, side=None):
        '''@todo:-implement the boundary extraction
        '''
        raise NotImplementedError

    def get_interior(self):
        '''@todo:-implement the boundary extraction
        '''
        raise NotImplementedError

    def get_ls_mask(self, ls_mask_function):
        '''
        Return a boolean array indicating masked nodes.
        '''
        return self.geo_grid.get_ls_mask(ls_mask_function)

    def get_intersected_elems(self, ls_function, ls_limits=None):
        '''
        Return elems intersected by specified domain.

        Requirements on the e_domain
        - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        The e_domain must be at most n - 1 dimensionality of the grid.

        Two methods must be supported by the e_domain
        1 ) The e_domain must be discretizable within the current grid. That means
           it should return arrays with points on grid lines that intersect the
           e_domain's boundaries
        2) With the list of intersections, it is possible to identify the intersected
           elements.
        These two methods would avoid a full search through all the grid points.

        Applicable:
        -----------
        Boundary conditions including the shape function coefficient
        Local - element enrirchments (XFEM)

        Method result
        -------------
        The method returns the

        '''
        return self.geo_grid.get_intersected_cells(ls_function, ls_limits)

    def get_negative_elems(self, ls_function):
        return self.geo_grid.get_negative_cells(ls_function)

    def get_affected_elems(self, e_domain):
        '''
        3) It should have a notion of inside / outside to decide whether or not
           a point is to be included or not.
        '''

    def get_enclosed_nodes(self, e_domain):
        '''
        Get elements that are inside of the e_domain.
        The elements are not intersected by the boundaries of the e_domain.

        The e_domain must be at least the same dimension as the grid.
        The e_domain must have an operator inside/outside to decide about
        whether or not a point is included.

        The method returns an array of element numbers within the specified e_domain.
        '''

    def get_subgrid(self, bounding_box):
        '''
        Return subgrid with base_node number
        '''

    def get_ls_value(self, X_pnt):
        # @TODO: make it work for 3d
        return self.ls_function.level_set_fn(X_pnt[0], X_pnt[1])

    _grids = Property(
        depends_on='fets_eval.dof_r,fets_eval.geo_r,shape+,coord_min+,coord_max+,fets_eval.n_nodal_dofs,dof_offset,geo_transform')

    @cached_property
    def _get__grids(self):
        '''Regenerate grids based on the specification
        '''
        # Check if the specifiers for both grids are identical
        #
        cell_grid = CellGrid(
            grid_cell_spec=self.dof_grid_spec,
            geo_transform=self.geo_transform,
            shape=self.shape,
            coord_min=self.coord_min,
            coord_max=self.coord_max
        )
        _xdof_grid = DofCellGrid(
            cell_grid=cell_grid,
            n_nodal_dofs=self.n_nodal_dofs,
            dof_offset=self.dof_offset
        )

        # if self.dof_r.all() != self.geo_r.all():
        #
        # @todo check whether this is universally valid (tolerance of coordinate values)
        #
        if not array_equal(self.dof_r, self.geo_r):
            #        if (self.dof_r.shape[0] != self.geo_r.shape[0])or\
            #           ((self.dof_r.shape == self.geo_r.shape) and\
            #            linalg.norm(self.dof_r - self.geo_r) > 1.e-3) :
            # If the geometry grid differs from the dof_grid specifier
            # construct a separate cell grid, otherwise reuse the dof_grid
            cell_grid = CellGrid(grid_cell_spec=self.geo_grid_spec,
                                 geo_transform=self.geo_transform,
                                 shape=self.shape,
                                 coord_min=self.coord_min,
                                 coord_max=self.coord_max)
        _xgeo_grid = GeoCellGrid(cell_grid=cell_grid)
        return (_xdof_grid, _xgeo_grid)


class FECellView(CellView):
    geo_view = Instance(GeoCellView)

    def _geo_view_default(self):
        return GeoCellView()

    dof_view = Instance(DofCellView)

    def _dof_view_default(self):
        return DofCellView()

    @on_trait_change('cell_grid')
    def _reset_view_links(self):
        self.geo_view.cell_grid = self.cell_grid.geo_grid
        self.dof_view.cell_grid = self.cell_grid.dof_grid

    def set_cell_traits(self):
        self.dof_view.cell_idx = self.cell_idx
        self.dof_view.set_cell_traits()
        self.geo_view.cell_idx = self.cell_idx
        self.geo_view.set_cell_traits()

    def redraw(self):
        self.dof_view.redraw()
        self.geo_view.redraw()





from ibvpy.bcond.bcond_mngr import \
    BCondMngr
from ibvpy.mathkit.linalg.sys_mtx_assembly import \
    SysMtxAssembly

import numpy as np
import traits.api as tr

from .domain_state import DomainState
from .i_tstep import ITStep
from .tloop_implicit import TLoopImplicit
from .tstep import TStep
from .domain_state_container import DomainStateContainer


@tr.provides(ITStep)
class TStepBC(TStep):

    tloop_type = tr.Type(TLoopImplicit)

    primary_var_changed = tr.Event

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.fe_domain

    #=========================================================================
    # Time and state
    #=========================================================================

    t_n = tr.Float(0.0)
    '''Fundamental state time used for time dependent essential BC'
    '''
    U_n = tr.Property(tr.Array(np.float64),
                      depends_on='model_structure_changed')
    '''Fundamental value of the primary (control variable)
    '''
    @tr.cached_property
    def _get_U_n(self):
        return np.zeros_like(self.U_k)

    '''Current fundamental value of the primary variable.
    '''
    U_k = tr.Property(tr.Array(np.float64),
                      depends_on='model_structure_changed')
    '''Fundamental value of the primary (control variable)
    '''
    @tr.cached_property
    def _get_U_k(self):

        U_var_shape = self.fe_domain.U_var_shape
        return np.zeros(U_var_shape, dtype=np.float64).flatten()

    model_structure_changed = tr.Event

    bc = tr.List
    r'''Boundary conditions
    '''
    record = tr.Dict
    r'''Recorded variables
    '''

    # Boundary condition manager
    #
    bcond_mngr = tr.Property(tr.Instance(BCondMngr),
                             depends_on='bc,bc_items,model_structure_changed')

    @tr.cached_property
    def _get_bcond_mngr(self):
        return BCondMngr(bcond_list=self.bc)

    def init_state(self):
        '''Initialize state.
        '''
        self.t_n = 0.0
        self.t_n1 = 0.0
        self.model_structure_changed = True

    step_flag = tr.Enum('predictor', 'corrector')
    '''Step flag to control the inclusion of essential BC'
    '''
    @tr.on_trait_change('model_structure_changed')
    def _reset(self):
        self.step_flag = 'predictor'

    K = tr.Property(tr.Instance(SysMtxAssembly),
        depends_on='model_structure_changed'
    )
    '''System matrix with registered essencial boundary conditions.
    '''
    @tr.cached_property
    def _get_K(self):
        K = SysMtxAssembly()
        self.bcond_mngr.setup(None)
        self.bcond_mngr.register(K)
        return K

    #=========================================================================
    # Spatial domain
    #=========================================================================
    domains = tr.List([])
    r'''Spatial domain represented by a finite element discretization.
    providing the kinematic mapping between the linear algebra (vector and
    matrix) and field representation of the primary variables.
    '''

    fe_domain = tr.Property(depends_on='model_structure_changed, state_changed')

    @tr.cached_property
    def _get_fe_domain(self):
        domains = [
            DomainState(tstep=self, xmodel=xmodel, tmodel=tmodel)
            for xmodel, tmodel in self.domains
        ]
        return DomainStateContainer(domains)

    corr_pred = tr.Property(depends_on='primary_var_changed,t_n1')

    @tr.cached_property
    def _get_corr_pred(self):
        self.K.reset_mtx()
        f_Eis, K_ks, dof_Es = np.array(
            [s.get_corr_pred(self.U_k, self.t_n, self.t_n1)
             for s in self.fe_domain],
        dtype=object).T
        self.K.sys_mtx_arrays = list(K_ks)  # improve
        F_ext = np.zeros_like(self.U_k)
        self.bcond_mngr.apply(
            self.step_flag, None, self.K, F_ext, self.t_n, self.t_n1
        )
        F_int = np.bincount(
            np.hstack(np.hstack(dof_Es)),
            weights=np.hstack(np.hstack(f_Eis))
        )
        R = F_ext - F_int
        self.K.apply_constraints(R)
        if self.debug:
            print('t_n1', self.t_n1)
            print('U_k\n', self.U_k)
            print('F_int\n', F_int)
            print('F_ext\n', F_ext)
        return R, self.K, F_int

    def make_iter(self):
        '''Perform a single iteration
        '''
        d_U_k, pos_def = self.K.solve(check_pos_def=True)
        if self.debug:
            print('positive definite', pos_def)
        self.U_k[:] += d_U_k
        self.primary_var_changed = True
        self.step_flag = 'corrector'

    def make_incr(self):
        '''Update the control, primary and state variables..
        '''
        self.U_n[:] = self.U_k[:]
        states = [d.record_state() for d in self.fe_domain]
        self.hist.record_timestep(self.t_n1, self.U_k, self.F_k, states)
        self.t_n = self.t_n1
        self.step_flag = 'predictor'

    def __str__(self):
        s = '\nt_n: %g, t_n1: %g' % (self.t_n, self.t_n1)
        s += '\nU_n' + str(self.U_n)
        s += '\nU_k' + str(self.U_k)
        return s

    R_norm = tr.Property

    def _get_R_norm(self):
        R = self.R
        return np.sqrt(np.einsum('...i,...i', R, R))

    R = tr.Property

    def _get_R(self):
        R, _, _ = self.corr_pred
        return R.flatten()

    dR = tr.Property

    def _get_dR(self):
        _, dR, _ = self.corr_pred
        return dR

    F_k = tr.Property

    def _get_F_k(self):
        _, _, F_k = self.corr_pred
        return F_k



'''
@author: rch
'''

from traits.api import \
    Int, Float, Type, Bool
from .tloop import \
    TLoop


class TLoopImplicit(TLoop):
    '''Time loop with implicit time stepping controlling the newton
    based algorithms.
    '''

    k_max = Int(100, enter_set=True, auto_set=False)

    acc = Float(1e-4, enter_set=True, auto_set=False)

    verbose = Bool(False, enter_set=True, auto_set=False)

    def eval(self):
        t_n1 = self.tline.val
        t_max = self.tline.max
        dt = self.tline.step

        if self.verbose:
            print('t:', end='')

        while t_n1 <= (t_max + 1e-8):
            if self.verbose:
                print('\t%5.2f' % t_n1, end='')

            k = 0
            self.tstep.t_n1 = t_n1
            # run the iteration loop
            while (k < self.k_max) and not self.user_wants_abort:
                if self.tstep.R_norm < self.acc:
                    if self.verbose:
                        print('(%g), ' % k, end='\n')
                    break
                try:
                    self.tstep.make_iter()
                except RuntimeError as e:
                    raise(e)
                k += 1
            else:  # handle unfinished iteration loop
                if k >= self.k_max:  # add step size reduction
                    # no success abort the simulation
                    self.restart = True
                    print('')
                    raise StopIteration('Warning: '
                                        'convergence not reached in %g iterations' % k)
                return
            # accept the time step and record the state in history
            self.tstep.make_incr()
            # update the line - launches notifiers to subscribers
            self.tline.val = min(t_n1, self.tline.max)
            # set a new target time
            t_n1 += dt

        if self.verbose:
            print('')

        return



'''
Created on 15.02.2018

@author: abaktheer

Microplane damage model 2D - Jirasek [1999]
'''

from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from numpy import \
    array, einsum, identity, sqrt
from traits.api import \
    Constant, \
    Property, cached_property

from bmcs_utils.api import Float

import numpy as np
import traits.api as tr


class MATS3DMplDamageEEQ(MATS3DEval):
    # To use the model directly in the simulator specify the
    # time stepping classes

    epsilon_0 = Float(59.0e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    epsilon_f = Float(250.0e-6,
                      label="a",
                      desc="Lateral pressure coefficient",
                      enter_set=True,
                      auto_set=False)

    c_T = Float(0.01,
                label="a",
                desc="Lateral pressure coefficient",
                enter_set=True,
                auto_set=False)

    #=========================================================================
    # Configurational parameters
    #=========================================================================
    state_var_shapes = tr.Property(tr.Dict(), depends_on='n_mp')
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''
    @cached_property
    def _get_state_var_shapes(self):
        return {'kappa_n': (self.n_mp,),
                'omega_n': (self.n_mp,)}

    U_var_shape = (6,)
    '''Shape of the primary variable required by the TStepState.
    '''

    node_name = 'Desmorat model'

    tree_node_list = tr.List([])

    #=========================================================================
    # Evaluation - get the corrector and predictor
    #=========================================================================

    def get_corr_pred(self, eps_ab, tn1, kappa_n, omega_n):

        self._update_state_variables(eps_ab, kappa_n, omega_n)
        #----------------------------------------------------------------------
        # if the regularization using the crack-band concept is on calculate the
        # effective element length in the direction of principle strains
        #----------------------------------------------------------------------
        # if self.regularization:
        #    h = self.get_regularizing_length(sctx, eps_app_eng)
        #    self.phi_fn.h = h

        #------------------------------------------------------------------
        # Damage tensor (2th order):
        #------------------------------------------------------------------
        phi_ab = self._get_phi_ab(kappa_n)

        #------------------------------------------------------------------
        # Damage tensor (4th order) using product- or sum-type symmetrization:
        #------------------------------------------------------------------
        beta_abcd = self._get_beta_abcd(phi_ab)

        #------------------------------------------------------------------
        # Damaged stiffness tensor calculated based on the damage tensor beta4:
        #------------------------------------------------------------------
        D_ijab = einsum(
            '...ijab, abef, ...cdef -> ...ijcd',
            beta_abcd, self.D_abef, beta_abcd
        )

        sig_ab = einsum(
            '...abef,...ef -> ...ab',
            D_ijab, eps_ab
        )

        return sig_ab, D_ijab

    #=========================================================================
    # MICROPLANE-Kinematic constraints
    #=========================================================================

    _MPNN = Property(depends_on='n_mp')
    r'''Get the dyadic product of the microplane normals
    '''
    @cached_property
    def _get__MPNN(self):
        # dyadic product of the microplane normals

        MPNN_nij = einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    _MPTT = Property(depends_on='n_mp')
    r'''Get the third order tangential tensor (operator) for each microplane
    '''
    @cached_property
    def _get__MPTT(self):
        # Third order tangential tensor for each microplane
        delta = identity(3)
        MPTT_nijr = 0.5 * (
            einsum('ni,jr -> nijr', self._MPN, delta) +
            einsum('nj,ir -> njir', self._MPN, delta) - 2.0 *
            einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN)
        )
        return MPTT_nijr

    def _get_e_na(self, eps_ab):
        r'''
        Projection of apparent strain onto the individual microplanes
        '''
        e_ni = einsum(
            'nb,...ba->...na',
            self._MPN, eps_ab
        )
        return e_ni

    def _get_e_N_n(self, e_na):
        r'''
        Get the normal strain array for each microplane
        '''
        e_N_n = einsum(
            '...na, na->...n',
            e_na, self._MPN
        )
        return e_N_n

    def _get_e_equiv_n(self, e_na):
        r'''
        Returns a list of the microplane equivalent strains
        based on the list of microplane strain vectors
        '''
        # magnitude of the normal strain vector for each microplane
        e_N_n = self._get_e_N_n(e_na)
        # positive part of the normal strain magnitude for each microplane
        e_N_pos_n = (np.abs(e_N_n) + e_N_n) / 2.0
        # normal strain vector for each microplane
        e_N_na = einsum('...n,ni -> ...ni', e_N_n, self._MPN)
        # tangent strain ratio
        c_T = self.c_T
        # tangential strain vector for each microplane
        e_T_na = e_na - e_N_na
        # squared tangential strain vector for each microplane
        e_TT_n = einsum('...ni,...ni -> ...n', e_T_na, e_T_na)
        # equivalent strain for each microplane
        e_equiv_n = sqrt(e_N_pos_n * e_N_pos_n + c_T * e_TT_n)
        return e_equiv_n

    def _update_state_variables(self, eps_ab, kappa_n, omega_n):
        e_na = self._get_e_na(eps_ab)
        eps_eq_n = self._get_e_equiv_n(e_na)
        f_trial_n = eps_eq_n - self.epsilon_0
        I = np.where(f_trial_n > 0)
        k_n = np.max(np.array([kappa_n[I], eps_eq_n[I]]), axis=0)
        kappa_n[I] = k_n
        omega_n[I] = self._get_omega(k_n)

    def _get_omega(self, kappa_n):
        '''
        Return new value of damage parameter
        @param kappa:
        '''
        omega_n = np.zeros_like(kappa_n)
        epsilon_0 = self.epsilon_0
        epsilon_f = self.epsilon_f
        I = np.where(kappa_n >= epsilon_0)
        omega_n[I] = (
            1.0 - (
                epsilon_0 / kappa_n[I] *
                   np.exp(-1.0 * (kappa_n[I] - epsilon_0) /
                          (epsilon_f - epsilon_0))
                   )
        )
        return omega_n

    def _get_phi_ab(self, kappa_n):
        # Returns the 2nd order damage tensor 'phi_mtx'
        # scalar integrity factor for each microplane
        phi_n = np.sqrt(1.0 - self._get_omega(kappa_n))
        # print 'phi_Emn', phi_Emn[:, -1, :]
        # integration terms for each microplanes
        phi_ab = einsum('...n,n,nab->...ab', phi_n, self._MPW, self._MPNN)
        return phi_ab

    def _get_beta_abcd(self, phi_ab):
        '''
        Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
        (cf. [Jir99], Eq.(21))
        '''
        delta = identity(3)
        beta_ijkl = 0.25 * (
            einsum('...ik,jl->...ijkl', phi_ab, delta) +
            einsum('...il,jk->...ijkl', phi_ab, delta) +
            einsum('...jk,il->...ijkl', phi_ab, delta) +
            einsum('...jl,ik->...ijkl', phi_ab, delta)
        )
        return beta_ijkl

    #-----------------------------------------------
    # number of microplanes - currently fixed for 3D
    #-----------------------------------------------
    n_mp = Constant(28)

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return array([[.577350259, .577350259, .577350259],
                      [.577350259, .577350259, -.577350259],
                      [.577350259, -.577350259, .577350259],
                      [.577350259, -.577350259, -.577350259],
                      [.935113132, .250562787, .250562787],
                      [.935113132, .250562787, -.250562787],
                      [.935113132, -.250562787, .250562787],
                      [.935113132, -.250562787, -.250562787],
                      [.250562787, .935113132, .250562787],
                      [.250562787, .935113132, -.250562787],
                      [.250562787, -.935113132, .250562787],
                      [.250562787, -.935113132, -.250562787],
                      [.250562787, .250562787, .935113132],
                      [.250562787, .250562787, -.935113132],
                      [.250562787, -.250562787, .935113132],
                      [.250562787, -.250562787, -.935113132],
                      [.186156720, .694746614, .694746614],
                      [.186156720, .694746614, -.694746614],
                      [.186156720, -.694746614, .694746614],
                      [.186156720, -.694746614, -.694746614],
                      [.694746614, .186156720, .694746614],
                      [.694746614, .186156720, -.694746614],
                      [.694746614, -.186156720, .694746614],
                      [.694746614, -.186156720, -.694746614],
                      [.694746614, .694746614, .186156720],
                      [.694746614, .694746614, -.186156720],
                      [.694746614, -.694746614, .186156720],
                      [.694746614, -.694746614, -.186156720]])

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return array([.0160714276, .0160714276, .0160714276, .0160714276, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505]) * 6.0

    def _get_lame_params(self):
        la = self.E * self.nu / ((1. + self.nu) * (1. - 2. * self.nu))
        # second Lame parameter (shear modulus)
        mu = self.E / (2. + 2. * self.nu)
        return la, mu

    D_abef = tr.Property(tr.Array, depends_on='+input')

    @tr.cached_property
    def _get_D_abef(self):
        la = self._get_lame_params()[0]
        mu = self._get_lame_params()[1]
        delta = identity(3)
        D_abef = (einsum(',ij,kl->ijkl', la, delta, delta) +
                  einsum(',ik,jl->ijkl', mu, delta, delta) +
                  einsum(',il,jk->ijkl', mu, delta, delta))

        return D_abef

    #-----------------------------------------------------------
    # Response variables
    #-----------------------------------------------------------

    def get_phi_ab(self, eps_ab, tn1, kappa_n, omega_n):
        return self._get_phi_ab(kappa_n)

    def get_omega_ab(self, eps_ab, tn1, kappa_n, omega_n):
        return np.identity(3) - self._get_phi_ab(kappa_n)

    def get_max_omega(self, eps_Emab, t_n1, kappa_n, omega_n):
        return np.max(omega_n, axis=-1)

    def _get_var_dict(self):
        var_dict = super(MATS3DMplDamageEEQ, self)._get_var_dict()
        var_dict.update(
            max_omega=self.get_max_omega,
            phi_ab=self.get_phi_ab,
            omega_ab=self.get_omega_ab
        )
        return var_dict
